@@ -122,7 +122,7 @@ class SequenceTransferLearningMultiLabelWrapper(EncoderModelWrapper, TrainableMo
       print('Encoder Output Shape = ' + str(encoder_shape))
 
       # Input shape
-      input_tensor = self.get_input_tensors()
+      input_tensor = self.get_preprocessed_input_tensors()
       input_tensor  = Lambda(lambda x:x[:,0:encoder_shape[1]])(input_tensor) # Filter out to have matched input and output length
       input_shape = input_tensor.get_shape().as_list()
       print('Input Shape = ' + str(input_shape))
@@ -200,7 +200,7 @@ class SequenceTransferLearningMultiLabelWrapper(EncoderModelWrapper, TrainableMo
 
   # Callback activated when just before model being compile.
   # Keras model is passed so We can use it to load saved weight from checkpoint.
-  def on_before_compile(self, model):
+  def on_after_init(self, model):
     model.summary()
 
     # Load language model checkpoint
